# Author: Sreepathi Pai

#from cuda_occupancy import gpu_configs, get_resident_blocks, gpu_specs

# THREADS
GTID = 0
GBID = 1
BID = 2
TID = 3
START = 4
END = 5
THREADS_MPROC = 6
WAVE_ID = 7

# GPUSPEC
MPROC = 0
SMEM = 1

# KERNELCFG
NAME = 0
GRID_HEIGHT=1
GRID_WIDTH=2
BLOCK_X=3
BLOCK_Y=4
BLOCK_Z=5

# RESOURCES
STAT_SMEM = 0
CMEM = 1
LMEM = 2
NREGS = 3
DYN_SMEM = 4

def get_headers_as_string(headers):    
    out = []
    for (k, v) in headers.iteritems():
        if isinstance(v, list):
            vv = " ".join([str(v1) for v1 in v])
        else:
            vv = v

        out.append("##%s\n%s\n" % (k, vv))

    return "".join(out)

def read_header(f):

    # GPU Name
    # GPU Specs
    # Kernel Specs (grid, block)
    # Runtime Info (resident blocks)
    
    headers = {}
    
    expecting_header = True

    header = ""
    for l in f:
        l = l.strip()
        if l[0:2] == "##":
            assert expecting_header
            header = l[2:].strip().upper()
            expecting_header = False

            if header == "THREADS" or header == "BLOCKS":
                break            
        else:
            value = l
                     
            if header in ("GPU", "COMPUTECAPABILITY"):
                headers[header] = value
            elif header == "GPUSPECS":
                headers[header] = [int(k) for k in value.split()]
            elif header == "KERNELCFG":
                vs = value.split()
                headers[header] = [vs[0]] + [int(k) for k in vs[1:]]
            elif header == "RESIDENT":
                headers[header] = int(value)
            elif header == "RESOURCES":
                headers[header] = [int(k) for k in value.split()]
            elif header == "CORRECTION":
                headers[header] = [int(k) for k in value.split()]
            else:
                headers[header] = value
                
            expecting_header = True

    return (header, headers)

def add_headers(user_supplied, headers):
    if "COMPUTECAPABILITY" in headers and not "GPUSPECS" in headers:
        gc = gpu_configs[headers["COMPUTECAPABILITY"]]
        headers["GPUSPECS"] = [0, 0]

        if "GPU" in headers:
            gs = gpu_specs[headers["GPU"]]
        else:
            for gs in gpu_specs:
                if gs['version'] == headers["COMPUTECAPABILITY"]:
                    break
                
        headers["GPUSPECS"][MPROC] = gs["mp"]
        headers["GPUSPECS"][SMEM] = gc["total_sm"]

    if "RESIDENT" in user_supplied:
        headers["RESIDENT"] = user_supplied["RESIDENT"]

    if "RESOURCES" in headers and not "RESIDENT" in headers:
        resident = get_resident_blocks(headers["KERNELCFG"][BLOCK_X],
                                       headers["KERNELCFG"][BLOCK_Y],
                                       headers["KERNELCFG"][BLOCK_Z],
                                       headers["RESOURCES"][NREGS],
                                       headers["RESOURCES"][DYN_SMEM],
                                       headers["RESOURCES"][STAT_SMEM],
                                       gpu_configs[headers["COMPUTECAPABILITY"]])

        headers["RESIDENT"] = resident


def estimate_wavelength(data, headers):
    if "RESIDENT" in headers:
        wavelength = headers["GPUSPECS"][MPROC] * headers["RESIDENT"]
        #print "wavelength: ", wavelength

        return wavelength
    else:
        first = data[0]

        k = 0
        for d in data:
            if d[START] > first[END]:
                break
            
            k = k + 1
            
        #print "estimated wavelength: ", k
        
        return k

def read_threads(f, kernelcfg, read_threads = set([0]), start_block = 0, max_blocks = 256):
    # thread_id, thread_start, thread_end, [mproc]

    name, grid_height, grid_width, block_x, block_y, block_z = kernelcfg

    last_gbid = -1
    blocks = 0

    for l in f:
        if l[0] == "#":
            continue

        lf = l.strip().split(' ', 4)

        if len(lf) == 3:
            gtid, ts, te = [int(k) for k in lf]
            mproc = -1
            wave_id = -1
        elif len(lf) == 4:
            gtid, ts, te, mproc = [int(k) for k in lf]
            wave_id = -1
        elif len(lf) == 5:
            gtid, ts, te, mproc, wave_id = [int(k) for k in lf]

        gbid = (gtid / (block_x * block_y * block_z))

        if last_gbid != gbid:
            last_gbid = gbid
            blocks = blocks + 1
            if blocks > start_block + max_blocks:
                break

        if gbid >= start_block:
            # tid = tid.x + tid.y * (bdim.x) + tid.z * (bdim.x * bdim.y)
            tid = gtid % (block_x * block_y * block_z)
            tid_z = tid / (block_x * block_y)
            tid_y = (tid % (block_x * block_y)) / block_x
            tid_x = (tid - tid_y * block_x - tid_z * block_x * block_y) 

            # block_id = bid.x + bid.y * gdim.x
            bid_y = gbid / (grid_width)
            bid_x = gbid % (grid_width)
            bid_z = 0

            yield (gtid, gbid, (bid_x,bid_y), (tid_x, tid_y, tid_z), ts, te, mproc, wave_id)

def read_blocks(f):
    pass

def read_threads_as_block(f, kernelcfg, **kwargs):

    last_block = None
    block_start = 0
    block_end = 0
    
    for y in read_threads(f, kernelcfg, **kwargs):
        if last_block:
            if last_block[GBID] != y[GBID]:
                yield((-1, last_block[GBID], last_block[BID], (-1,-1,-1), block_start, block_end, last_block[THREADS_MPROC], last_block[WAVE_ID]))

                block_start = y[START]
                block_end = y[END]

            block_start = min(block_start, y[START])
            block_end = max(block_end, y[END])
        else:
            block_start = y[START]
            block_end = y[END]
            last_block = y
        
        last_block = y

    #print block_start, block_end, last_block

    yield((-1, last_block[GBID], last_block[BID], (-1,-1,-1), block_start, block_end, last_block[THREADS_MPROC], last_block[WAVE_ID]))
