/* -*- mode: c++;  compile-command: nvcc -arch sm_20 gpc_to_mp.cu -o gpc_to_mp -*- */

#include <cuda.h>
#include <stdio.h>
#include <assert.h>

#  define CUDA_SAFE_CALL_NO_SYNC( call) {                                    \
    cudaError err = call;                                                    \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
                __FILE__, __LINE__, cudaGetErrorString( err) );              \
        exit(EXIT_FAILURE);                                                  \
    } }

#  define CUDA_SAFE_CALL( call)     CUDA_SAFE_CALL_NO_SYNC(call);

extern __shared__ uint smoutput[];

// from http://forums.nvidia.com/index.php?showtopic=186669
__device__ uint get_smid(void) {
     uint ret;
     asm("mov.u32 %0, %smid;" : "=r"(ret) );
     return ret;
}

__global__ void testfn(clock_t *start_times, clock_t *end_times, uint *smids)
{
  int tid = blockIdx.y * gridDim.x + blockIdx.x;
  tid *= blockDim.x * blockDim.y * blockDim.z;
  tid += threadIdx.z * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;

  start_times[tid] = clock64();

  uint m_z = tid;
  uint m_w = tid;
  uint rnd;

  uint work_iter;
  work_iter = 10000;

  for(int i =0; i < work_iter;i++)
    {
      m_z = 36969 * (m_z & 65535) + (m_z >> 16);
      m_w = 18000 * (m_w & 65535) + (m_w >> 16);
      rnd = (m_z << 16) + m_w;     
    }

  smoutput[0] = rnd;

  smids[tid] = get_smid();
  
  end_times[tid] = clock64();
}

void identify_clusters_20(uint smids[], clock_t start_times[], clock_t end_times[], cudaDeviceProp *di) {
  if(!(di->multiProcessorCount == 14 || di->multiProcessorCount == 15)) {
    fprintf(stderr, "WARNING: Unknown number of multiprocessors. Possibly wrong clustering below.\n");
  }

  int mp_per_cluster[] = {4, 4, 3, 3};
  int nclusters = 4;

  if(di->multiProcessorCount == 15) mp_per_cluster[2] = 4;

  for(int i = 0, k = 0; i < nclusters; i++) {
    for(int j = 0; j < mp_per_cluster[i]; j++, k++) {
      printf("%d ", smids[k]);
    }

    printf("\n");
  }  
}

int main(int argc, char *argv[])
{
  int BLOCKS=64, GRID_HEIGHT, GRID_WIDTH;
  int THREADS=1;
  int resident_blocks=1;
  struct cudaDeviceProp devinfo;;
  cudaError_t cerr;
  int debug = 0;

  if((cerr = cudaGetDeviceProperties(&devinfo, 0)) != cudaSuccess) {
    fprintf(stderr, "ERROR: Failed to get cuda device information (code: %d).\n", cerr);
    exit(1);
  } 	   
  
  fprintf(stderr, "GPU: %s, MP: %d, CC: %d.%d\n", devinfo.name, devinfo.multiProcessorCount, devinfo.major, devinfo.minor);
  
  if(devinfo.major == 2 && devinfo.minor == 0) {
    GRID_WIDTH = devinfo.multiProcessorCount * 2;
    GRID_HEIGHT = 1;
  } else {
    fprintf(stderr, "Don't know how to identify clusters for this compute capability!\n");
    exit(1);
  }
  
  BLOCKS = GRID_HEIGHT * GRID_WIDTH;
  
  dim3 threads(THREADS);

  clock_t start_times[BLOCKS*THREADS];
  clock_t end_times[BLOCKS*THREADS];
  uint smids[BLOCKS*THREADS + 1];

  void *start_times_d;
  void *end_times_d; 
  void *smids_d;

  CUDA_SAFE_CALL(cudaMalloc(&start_times_d, BLOCKS*THREADS*sizeof(clock_t)));
  CUDA_SAFE_CALL(cudaMalloc(&end_times_d, BLOCKS*THREADS*sizeof(clock_t)));
  CUDA_SAFE_CALL(cudaMalloc(&smids_d,(BLOCKS*THREADS+1)*sizeof(uint)));

  CUDA_SAFE_CALL(cudaMemset(start_times_d, 0, BLOCKS*THREADS*sizeof(clock_t)));
  CUDA_SAFE_CALL(cudaMemset(end_times_d, 0,BLOCKS*THREADS*sizeof(clock_t)));
  CUDA_SAFE_CALL(cudaMemset(smids_d, 0,(BLOCKS*THREADS+1)*sizeof(uint)));

  dim3 gd(GRID_WIDTH, GRID_HEIGHT);

  testfn<<<gd,THREADS,(devinfo.sharedMemPerBlock/(resident_blocks+1))+1>>>((clock_t *) start_times_d, 
									   (clock_t *) end_times_d,
									   (uint *) smids_d);

  CUDA_SAFE_CALL(cudaMemcpy(start_times, start_times_d, BLOCKS*THREADS*sizeof(clock_t), cudaMemcpyDeviceToHost));
  CUDA_SAFE_CALL(cudaMemcpy(end_times, end_times_d, BLOCKS*THREADS*sizeof(clock_t), cudaMemcpyDeviceToHost));
  CUDA_SAFE_CALL(cudaMemcpy(smids, smids_d, (BLOCKS*THREADS+1)*sizeof(uint), cudaMemcpyDeviceToHost));

  if(devinfo.major == 2 && devinfo.minor == 0) {
    identify_clusters_20(smids, start_times, end_times, &devinfo);
  }

  if(debug) {
    printf("##GPU\n%s\n", devinfo.name);
    printf("##GPUSPECS\n%d %d\n", devinfo.multiProcessorCount, devinfo.sharedMemPerBlock);
    printf("##COMPUTECAPABILITY\n%d %d\n", devinfo.major, devinfo.minor);
    printf("##KERNELCFG\n%s %d %d %d %d %d\n", "schedule_block_syn", GRID_HEIGHT, GRID_WIDTH, THREADS, 1, 1);
    printf("##RESIDENT\n%d\n", resident_blocks);

    printf("##THREADS\n");
    for(int i = 0; i < BLOCKS*THREADS; i++)
      {
	printf("%d %lu %lu %u\n", i, start_times[i], end_times[i], smids[i]);
      }
  }
}
