#!/usr/bin/env python

# Author: Sreepathi Pai

import sys

class KernelGrid(object):
    def __init__(self, gridX, gridY, residency):
        self.gridX = gridX
        self.gridY = gridY
        self.residency = residency
        self.blocks = gridX * gridY

class Fermi(object):
    def __init__(self):
        self.mp_per_cluster = [len(self.cluster_to_mp_map[i]) for i in range(len(self.cluster_to_mp_map))]
        self.mp_to_cluster_map = {}

        for k, v  in self.cluster_to_mp_map.iteritems():
            self.mp_to_cluster_map.update(dict([(vv, k) for vv in v]))

    def pattern_2d(self, width, height):
        max_x = width / 2
        max_y = height / 2
        x = 0
        y = 0

        for i in range(0, max_y):
            if i % 2 == 0:
                for j in range(0, max_x):
                    # |_|
                    yield ( x, y)
                    yield ( x, y+1)
                    yield ( x+1, y+1)
                    yield ( x+1, y)

                    x = x + 2
            else:
                for j in range(0, max_x):
                    #  _
                    # | |
                    yield ( x, y)
                    yield ( x, y - 1)
                    yield ( x - 1, y - 1)
                    yield ( x - 1, y)

                    x = x - 2

            if width % 2 == 1:
                if i % 2 == 0:
                    # |
                    # |
                    # |
                    #<|
                    ctr = 0
                    while y < height and ctr < 4:            
                        yield (x, y)
                        y = y + 1
                        ctr += 1

                    x = x - 1
                    y = y - 1
                else:
                    #  _
                    # | |
                    # v
                    x = 0
                    y = y + 1
            else:
                if i % 2 == 0:
                    # |_|
                    #  _ 
                    # | |
                    #   ^
                    x = width - 1
                    y = y + 3
                else:
                    #  _
                    # | |
                    # v
                    x = 0
                    y = y + 1

        if height % 2 == 1:
            if y >= height:
                y = height - 1

            if max_y % 2 == 0:
                while x < width:
                    yield (x, y)
                    x = x + 1
            else:
                while x >= 0:
                    yield (x, y)
                    x = x - 1

    def sched_prio_residency(self, kg):
        total_clusters = len(self.mp_per_cluster)
        total_mp = sum(self.mp_per_cluster)

        firstwave = min(kg.blocks, total_mp * kg.residency)
        prio_cluster = [kg.residency * mpc for mpc in self.mp_per_cluster]

        dispatched = 0
        remaining = kg.blocks

        while dispatched < firstwave:
            max_prio = max(prio_cluster)
            max_cluster = prio_cluster.index(max_prio)

            if remaining > total_mp:
                dispatch = self.mp_per_cluster[max_cluster]
            else:
                dispatch = 1
            
            yield (max_cluster, dispatch, prio_cluster[max_cluster])

            dispatched += dispatch
            remaining -= dispatch
            prio_cluster[max_cluster] -= dispatch
        
    def sched_initial(self, kg):
        cluster_to_mp = [0, 0, 0, 0]

        def next_mp_in_cluster(cl):
            mp = cluster_to_mp[cl]
            cluster_to_mp[cl] = (cluster_to_mp[cl] + 1) % self.mp_per_cluster[cl]

            return mp
            
        k = 0
        blocks = iter(self.pattern_2d(kg.gridX, kg.gridY))
        for cluster, dispatch, priority in self.sched_prio_residency(kg):
            for i in range(0, dispatch):
                blkid = next(blocks)
                linblkid = blkid[0] + blkid[1] * kg.gridX
                yield (k, linblkid, blkid, cluster, self.cluster_to_mp_map[cluster][next_mp_in_cluster(cluster)], priority)
                k += 1

# These assignments may vary between cards
class Quadro6000(Fermi):
    cluster_to_mp_quadro_6000 = {0: [0, 4, 8, 12],
                                 3: [3, 7, 11, 13],
                                 1: [1, 5, 9],
                                 2: [2, 6, 10]
                                 }

    cluster_to_mp_map = cluster_to_mp_quadro_6000

class GTX480(Fermi):
    cluster_to_mp_map = {0: [0, 4, 8, 12],
                         3: [3, 7, 11],
                         1: [1, 5, 9, 13],
                         2: [2, 6, 10, 14]
                         }

GPUS = {'Quadro 6000': Quadro6000,
        'GeForce GTX 480': GTX480}

if __name__ == "__main__":
    sgpus = sorted(GPUS.keys())

    if len(sys.argv) < 5:
        print "Usage: %s gridX gridY residency GPU-id" % (sys.argv[0],)
        print "  where GPU-id is one of"
        print "\t" + "\n\t".join(["%d: %s" % (i, n) for i, n in enumerate(sgpus)])

        sys.exit()

    gridX = int(sys.argv[1])
    gridY = int(sys.argv[2])
    residency = int(sys.argv[3])
    gpuid = int(sys.argv[4])

    kg = KernelGrid(gridX, gridY, residency)
    gpu = GPUS[sgpus[gpuid]]()

    print "Scheduling %dx%d blocks residency=%d on %s" % (gridX, gridY, residency, sgpus[gpuid])
    for x in gpu.sched_initial(kg):
        order, linblkid, blkid, cluster, sm, priority = x
        print order, linblkid, blkid, cluster, sm, priority

